import os
from skimage import io
import numpy as np
from typing import List, Tuple
import aicspylibczi

# ap = argparse.ArgumentParser(description="Extract tif from an image and count neurons and ganglia.")
# ap.add_argument('file_path',type=str,help="Path (relative or absolute) to target image")
# ap.add_argument('-o', '--out_dir', type=str, help="Root directory where output should be stored, default is base dir of the input image")
# ap.add_argument('-c', '--c_number', default=0, type=int, help="indexes (start at 0) of the channel contains the marker of interest. Default = 0.")


# ap.add_argument('-i', '--shape_indexes', default=None, type=int, help="index (start at 0) of the z-stack that needs to be extracted.")
# ap.add_argument('-t', '--tile_size', default=None, type=int, nargs=2, help="Tuple representing targetted tile size (X-Y). Example: `-t 2000 2000`. Default is no tiling behaviour")

# args = ap.parse_args()
# # if no out_dir is given, take the base dir of the input image
# if args.out_dir is None:
#     args.out_dir = os.path.dirname(args.file_path)

def getMostInFocusImage(image_array_list: List[np.array]) -> Tuple[np.array, int]:   
    """Gets most in focus image and its index from a list of image-arrays.

    Parameters
    ----------
    image_array_list : List[np.array]
        image_array_list

    Returns
    -------
    Tuple[np.array, int]

    """
    for image in image_array_list:
        # calculate edges in image
        edged = laplace(image)
        # Calculate stdev of edges
        stdev = np.std(edged)
        stdev_list.append(stdev)
    
    # Find largest stdev in list
    largest = max(stdev_list)
    # Fidn which index it is to link back to the original list
    index = stdev_list.index(largest)
    return image_array_list[index], index

def readImage(file_path: str) -> Tuple[np.array, str]:
    """Reads in image-file and returns both the image and it's file type.

    Parameters
    ----------
    file_path : str
        file_path

    Returns
    -------
    Tuple[np.array, str]

    """
    if args_file_path.lower().endswith((".tif", ".tiff", "nd2")):
        image = io.imread(args_file_path)
        img_type = "tif"
    if args_file_path.lower().endswith(".czi"):
        image = aicspylibczi.CziFile(args_file_path) 
        img_type = "czi"
        
    return image, img_type


def extractMostInFocusZstack(image: np.ndarray, z_shape_index:int = -1) -> np.ndarray:
    """Extract the most in focus z-stack of an image.

    Parameters
    ----------
    image : np.ndarray
        image
    z_shape_index : int
        index that indicates which position in the image.shape contains the z-axis.

    Returns
    -------
    np.ndarray

    """
    if not len(image.shape) > 2:
        raise ValueError(f"Image shape expected to be larger than 2, image shape is {image.shape}")
    for i in range(image.shape[z_shape_index]):
        img_list.append(np.take(image, i, z_shape_index))
    mostInFocusImage = getMostInFocusImage(img_list)
    return mostInFocusImage



def extractSingleImages(image: np.ndarray, indexes_dict: dict, image_type: str, filename: str):
    """Extract from multi-dimensional input images all singe-stack images indicated by the indexes dict.

    Parameters
    ----------
    image : np.ndarray
        multi-dimensional input image
    indexes_dict : dict
        dict where keys are the shape-indexes and their values are indexes of which channels in that axis to extract. The combination of all key-value pairs will be extracted into single-channel images
    image_type : str
        image_type
    filename : str
        filename
    """
    if image_type == "tif":
        for shape_index, channel_indexes in indexes_dict.items():
            if isinstance(channel_indexes, int): # if it's only one channel, it'ss give an iterable error, so we gotta make it a list
                channel_indexes = [channel_indexes] 
            for channel_index in channel_indexes:
                if shape_index == None:
                    extracted_image = np.take(image, channel_index, axis=-1)
                else:
                    extracted_image = np.take(image, channel_index, axis=shape_index)
                io.imsave(f"{os.path.splitext(filename)[0]}_i{shape_index}_c{channel_index}.tif", extracted_image)
            
    elif image_type == "czi":
        for shape_index, channel_indexes in indexes_dict.items():
            # now we assume the shape indexes are letters referring to czi stuff
            if isinstance(channel_indexes, int): # if it's only one channel, it'ss give an iterable error, so we gotta make it a list
                channel_indexes = [channel_indexes] 
            for channel_index in channel_indexes:
                extracted_image,shape = image.read_image(shape_index=channel_index)
                extracted_image = extracted_image[0,0,0,0,0,:,:]
                io.imsave(f"{os.path.splitext(filename)[0]}_i{shape_index}_c{channel_index}.tif", extracted_image )

# def parseIndexes() #TODO



import numpy as np
from skimage import io
from typing import List


def maxIPstack(img_list: List[np.array]) -> np.array:
    """Takes maximum intensity projection of a list of images

    Parameters
    ----------
    img_list : List[np.array]
        list of images

    Returns
    -------
    np.array

    """
    parsed_list = img_list
    parsed_list = [img if isinstance(img, np.ndarray) else io.imread(img) for img in img_list]

    maxIP = np.maximum.reduce(parsed_list)
    return maxIP




import numpy as np
from skimage import io
from PIL import Image
from matplotlib import pyplot as plt
import re
from typing import Tuple, List
import cv2

def calculateTileGridStatistics(tile_grid_shape, tile_size_x: int, tile_size_y: int, form="row_by_row", direction="left_to_right") -> Tuple[int]:
    """Calculates all necessary grid statistics based on tile shape and size

    Parameters
    ----------
    tile_grid_shape : Tuple[int, int]
        Tuple of int representing the number of tiles that fit in the x and y dimensions of the original image respectively.
    tile_size_x : int
        Number of pixels that in the x dimension of the original image
    tile_size_y : int
        Number of pixels that in the y dimension of the original image
    form : str
        string representing how the tiles are layed out, choose either 'row_by_row' or 'col_by_col'
    direction : str
        direction in which the tiling occurs, choose either 'left_to_right' or 'right_to_left'

    Returns
    -------
    [int] total_n_tiles
       Total number of tiles in the original image
    [ndarray] tile_grid_array
        An array representing the layout of the tiles
    [int] original_x
        length of the original image in the x dimension
    [int] original_y
        length of the original image in the y dimension
       
    """
    total_n_tiles = tile_grid_shape[0]*tile_grid_shape[1]
    # Create range list of the tiles
    total_tiles_list = list(range(1,total_n_tiles+1))
    if direction == "right_to_left":
        total_tiles_list = list(range(total_n_tiles,0, -1))

    # Reshape the range list into the tile grid of the original image.
    # We swap the elements of the grid because the rest of the pipeline sees x and y as horizontal vs vertical, but numpy sees it as an array, where x = vertical movement
    swapped_grid = (tile_grid_shape[1],tile_grid_shape[0])
    if form=="row_by_row":
        tile_grid_array = np.reshape(total_tiles_list, swapped_grid)
    if form=="col_by_col":
        tile_grid_array = np.reshape(total_tiles_list, swapped_grid)
        tile_grid_array = np.transpose(tile_grid_array)

    # Creating an empty array the size of an original image
    original_x = tile_grid_array.shape[1] * tile_size_x
    original_y = tile_grid_array.shape[0] * tile_size_y
    return total_n_tiles, tile_grid_array, original_x, original_y


def stitchImageList(image_path_list: List[str], tile_grid_shape: Tuple[int], tile_size_x: int, tile_size_y: int , imtype: str = "grey" ) -> np.array:
    """Stitch a list of images back together based on a stats describing its tiling grid

    Parameters
    ----------
    image_path_list : str
        list of paths that point to the tiles. The list should be ordered along the subsequent tiles.
    tile_grid_shape : tuple[int]
        tuple indicating the shape of the tiling grid: e.g.: (x-dim, y-dim) 
    tile_size_x : int
        size of the tiles in the x-dimension
    tile_size_y : int
        size of the tiles in the y-dimension
    imtype : str
        string depicting the type of image. Choose either grey or RGB. Default = grey

    Returns
    -------
    np.array

    """
    #  Now we create a list of the original path images, but sorted by tile number:
    # By first creating a dict where the key is the tile number, and the value is the path
    # Tile number is extracted by re.findalling on "tiled_d" and then extracting the number from that, i cast it to int to make the sorting work like it should
    tile_image_dict = {int(re.findall(r"\d+", re.findall(r"tiled_\d+", image_path)[0])[0]): image_path for image_path in image_path_list}
    sorted_tile_images = [value for (key, value) in sorted(tile_image_dict.items(), key=lambda x:x[0])]

    total_n_tiles, tile_grid_array, original_x, original_y = calculateTileGridStatistics(tile_grid_shape, tile_size_x, tile_size_y, form="col_by_col", direction="left_to_right")
    print(tile_grid_array)

    if imtype=="grey":
        new_im = Image.new('I;16', (original_x, original_y ))
    elif imtype=="RGB":
        new_im = Image.new('RGB', (original_x, original_y ))

    for index, image_path in enumerate(sorted_tile_images, 1):
        im = Image.open(image_path)
        row_location, col_location = np.where(tile_grid_array==index) # this returns rows and columns, NOT X and Y, which is the opposite
        x_offset = int(col_location)*im.size[0]
        y_offset = int(row_location)*im.size[1]
        new_im.paste(im, (x_offset,y_offset))

    numpy_image = np.asarray(new_im)
    return numpy_image



import setuptools
from pathlib import Path
import setuptools
import sys
min_version = (3, 8)

base_dir = Path(__file__).parent.resolve()
version_file = base_dir / "__version__.py"
readme_file = base_dir / "README.md"

with version_file.open() as f:
    exec(f.read())

with readme_file.open(encoding = "utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name = "pyxelperfect",
    version = __version__,
    author = "LMIB",
    author_email = "david.wouters@kuleuven.be",
    description = "Collection of python powered image-processing functions",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    keywords = "image-processing",
    url = "https://github.com/sifrimlab/pyxelperfect",
    project_urls = {
        "Bug Reports": "https://github.com/sifrimlab/pyxelperfect",
        "Source": "https://github.com/sifrimlab/pyxelperfect",
    },
    packages = setuptools.find_packages(),
    python_requires = '>={}'.format('.'.join(str(n) for n in min_version)),
    install_requires = [
        "numpy",
        "tqdm",
        "scikit-image",
        "opencv-python",
        "pandas",
        "aicspylibczi",
        "matplotlib",
        "Pillow",
        "tifffile",
    ],
    classifiers = [
        "Development Status :: 3 - Alpha",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
)



from skimage.io import imread_collection, imsave, imread
import numpy as np
from tifffile import imsave
from typing import List, Tuple
import glob
import os

def calculateOptimalLargestResolution(glob_pattern: str, target_tile_height: int, target_tile_width: int) -> List[int]: 
    """Calculates the optimal maximum resolution based on a set of images assuming that it will be used to tile the images into equal tiles of a certain size, given by the input variables.


    Parameters
    ----------
    glob_pattern : str
        glob pattern hitting all images to be taken into account for the calculation.
    target_tile_height : int
        target_tile_height
    target_tile_width : int
        target_tile_width

    Returns
    -------
    List[int]

    """
    images_array = np.array(imread_collection(glob_pattern), dtype = object)

    heights = []
    widths = []
    for i in range(len(images_array)):
        heights.append(images_array[i].shape[0])
        widths.append(images_array[i].shape[1])
    max_rows = max(heights) 
    max_columns = max(widths)

    ydiv = np.ceil(max_rows / target_tile_height)
    xdiv = np.ceil(max_columns / target_tile_width)

    target_full_rows = ydiv * target_tile_height
    target_full_columns = xdiv * target_tile_width

    return target_full_rows, target_full_columns, ydiv, xdiv

def padImage(image: np.array, target_full_rows: int, target_full_columns: int) -> np.ndarray:
    """Pads an image with black pixels up until a certain number of rows and columns

    Parameters
    ----------
    image : np.array
        image to be padded
    target_full_rows : int
        Number of rows to be achieved by padding
    target_full_columns : int
        Number of columns to be achieved by padding

    Returns
    -------
    np.ndarray

    """
    rowdiff = target_full_rows - image.shape[0]
    columndiff = target_full_columns - image.shape[1]
    padded_img = np.pad(image, ((0, rowdiff), (0, columndiff)))
    return padded_img

def tileImage(image: np.ndarray, ydiv: int, xdiv: int, image_prefix: str="image_tile_"):
    """Tile an image into smaller tiles based on divisions in x and y axes, and saving them to tif files.

    Parameters
    ----------
    image : np.ndarray
        image
    ydiv : int
        Nr of divisions to make in the y axis
    xdiv : int
        Nr of divisions to make in the x axis
    image_prefix : str
        prefix to add to the new filenames of the tiles. default = image_tile
    """
    if not image_prefix.endswith("_"):
        image_prefix += "_"

    temp_split = np.array_split(image, ydiv, axis = 0)
    # Item sublist part is just to unpack a list of lists into one list
    final_split = [item for sublist in [np.array_split(row, xdiv, axis = 1) for row in temp_split] for item in sublist]

    for i, img in enumerate(final_split, 1):
        imsave(f"{image_prefix}_tile{i}.tif", img)

def tile(glob_pattern: str, target_tile_width: int, target_tile_height: int) -> Tuple[int]:
    """Tile the images caught by the glob pattern by first padding them to a global image size needed to tile them all into the same tile size, given by the input values.

    Parameters
    ----------
    glob_pattern : str
        glob_pattern that catches all images to be tiled.
    target_tile_width : int
        Width of the tiles
    target_tile_height : int
        Height of the tiles

    Returns
    -------
    Tuple[int]

    """

    target_full_rows, target_full_columns, ydiv, xdiv = calculateOptimalLargestResolution(glob_pattern, target_tile_height, target_tile_width)
    padded_imgs = {}
    for image_path in glob.glob(glob_pattern):
        padded_imgs[os.path.splitext(os.path.basename(image_path))[0]] = padImage(image_path, int(target_full_rows), int(target_full_columns))

    for k, padded_img in padded_imgs.items():
        tileImage(padded_img, target_tile_height = target_full_rows, target_tile_width = target_full_columns, ydiv = ydiv, xdiv = xdiv, image_prefix = k)
    return xdiv, ydiv
